#!/usr/bin/python

import wx
import os
import commands
import sys
import gzip

from agar import *

from testtree import *
from rubrictree import *
from subtree import *
from uctree import *
from resulttree import *
from resultwin import *
from MailOptionsDialog import *
from MergeDialog import *
from gnumeripy import *

# Menu constants
ID_NEW = wx.NewId()
ID_SAVE = wx.NewId()
ID_SAVE_AS = wx.NewId()
ID_LOAD = wx.NewId()
ID_EXPORT_GRADES = wx.NewId()
ID_EXIT = wx.NewId()
ID_MERGE = wx.NewId()

ID_CONFIGURE = wx.NewId()

ID_POST_OCR_WIZ = wx.NewId()
ID_WRITTEN_WIZ = wx.NewId()

ID_RUN = wx.NewId()
ID_RUN_CHANGES = wx.NewId()
ID_SEND_MAILBACKS = wx.NewId()
ID_WRITE_MAILBACKS = wx.NewId()
ID_RUN_IN_ALL = wx.NewId()
ID_ADD_TO_ALL = wx.NewId()
ID_CONVERT = wx.NewId()
ID_DIAGNOSTIC = wx.NewId()

ID_PROJ_NAME = wx.NewId()
ID_MAIL_OPTIONS = wx.NewId()

ID_ABOUT = wx.NewId()
ID_HELP = wx.NewId()

class AgarFrame(wx.Frame):
    """This is the main class for the wxPython GUI."""

    def __init__(self, *args, **kwds):
        """Build the widgets, set up the menus, start up an instance
        of an Agar object (which does all the real work)"""

        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)

        self.filename = None

        if len(sys.argv) > 1 and sys.argv[1].endswith(".agar"):
            self.load(sys.argv[1])
        else:
            dlg = wx.TextEntryDialog(self,
                                     "Submission file mask:",
                                     "File mask",
                                     defaultValue="*.c,*.cc,*.C,*.cpp,*.h,"+\
                                                  "Makefile",
                                     style=wx.OK|wx.CANCEL)
            ret = dlg.ShowModal()
            val = str(dlg.GetValue())
            dlg.Destroy()

            if ret == wx.ID_CANCEL:
                sys.exit(0)            

            if len(sys.argv) > 1 and os.path.isdir(sys.argv[1]):
                sys.argv[1] = os.path.realpath(sys.argv[1])
                os.chdir(os.path.dirname(sys.argv[1]))
                self.a = Agar(os.path.basename(sys.argv[1]), val)
            else:
                self.onNew(val)
                if not hasattr(self, "a"):
                    sys.exit(0)

        self.progDlg = None
        self.aFrame = None

        self.addWidgets()
        self.doLayout()
        self.addHandlers()

        self.SetSize((800, 600))
        
        self.SetTitle("Agar - " + self.a.getAssignmentName())

    def addHandlers(self):
        """Set up event handlers"""
        wx.EVT_SIZE(self, self.onSize)

        # Menubar
        wx.EVT_MENU(self, ID_NEW, self.onNew)
        wx.EVT_MENU(self, ID_SAVE, self.onSave)
        wx.EVT_MENU(self, ID_SAVE_AS, self.onSaveAs)
        wx.EVT_MENU(self, ID_LOAD, self.onLoad)
        wx.EVT_MENU(self, ID_EXPORT_GRADES, self.onExport)
        wx.EVT_MENU(self, ID_MERGE, self.onMerge)
        wx.EVT_MENU(self, ID_EXIT, self.onExit)

        wx.EVT_MENU(self, ID_RUN, self.onRun)
        wx.EVT_MENU(self, ID_RUN_IN_ALL, self.onRunInAll)
        wx.EVT_MENU(self, ID_ADD_TO_ALL, self.onAddToAll)
        wx.EVT_MENU(self, ID_CONVERT, self.onConvertAll)
        wx.EVT_MENU(self, ID_SEND_MAILBACKS, self.onSendMailbacks)
        wx.EVT_MENU(self, ID_WRITE_MAILBACKS, self.onWriteMailbacks)

        wx.EVT_MENU(self, ID_PROJ_NAME, self.onSetProjName)
        wx.EVT_MENU(self, ID_MAIL_OPTIONS, self.onSetMailOptions)
        
    def addWidgets(self):
        """Set all the GUI widgets"""
        self.SetSize((800, 600))
        self.notebook = wx.Notebook(self)
        self.rubricPage = wx.Panel(self.notebook)
        self.rubricSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.tests = TestTree(self.a, self.rubricPage, -1,
                              style=wx.TR_HAS_BUTTONS|wx.TR_LINES_AT_ROOT|\
                                 wx.TR_DEFAULT_STYLE|wx.SUNKEN_BORDER)
        self.rubricSizer.Add(self.tests, 1, wx.EXPAND|wx.ALL, 3)
        self.rubric = RubricTree(self, self.a, self.rubricPage, -1,
                                 style=wx.TR_HAS_BUTTONS|wx.TR_LINES_AT_ROOT|\
                                 wx.TR_DEFAULT_STYLE|wx.SUNKEN_BORDER)
        self.rubricSizer.Add(self.rubric, 1, wx.EXPAND|wx.ALL, 3)
        self.rubricPage.SetSizer(self.rubricSizer)
        
        self.gradingPage = wx.Panel(self.notebook)
        self.gradingSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.subTree = SubTree(self, self.a, self.gradingPage, -1,
                               style=wx.TR_HAS_BUTTONS|\
                               wx.TR_LINES_AT_ROOT|wx.TR_DEFAULT_STYLE|\
                               wx.SUNKEN_BORDER)
        self.commentTree = UserCommentTree(self, self.a, 
                                           self.gradingPage, -1,
                                       style=wx.TR_HAS_BUTTONS|\
                                       wx.TR_LINES_AT_ROOT|wx.TR_DEFAULT_STYLE\
                                       |wx.SUNKEN_BORDER)
        self.resultTree = ResultTree(self, self.a, 
                                     self.gradingPage, -1,
                                     style=wx.TR_HAS_BUTTONS|\
                                     wx.TR_LINES_AT_ROOT|wx.TR_DEFAULT_STYLE|\
                                     wx.SUNKEN_BORDER)
        self.resultWin = ResultWin(self, self.a,
                                   self.gradingPage, -1,
                                   style=wx.HSCROLL|wx.TE_MULTILINE|\
                                   wx.TE_READONLY|wx.SUNKEN_BORDER)

        self.gradingSizer.Add(self.subTree, 2, wx.EXPAND|wx.ALL, 3)
        self.gradingSizer.Add(self.commentTree, 2, wx.EXPAND|wx.ALL, 3)
        self.gradingVSizer = wx.BoxSizer(wx.VERTICAL)
        self.gradingVSizer.Add(self.resultTree, 2, wx.EXPAND|wx.ALL, 3)
        self.gradingVSizer.Add(self.resultWin, 1, wx.EXPAND|wx.ALL, 3)
        self.gradingSizer.Add(self.gradingVSizer, 3, wx.EXPAND|wx.ALL, 3)
        self.gradingPage.SetSizer(self.gradingSizer)
        
        self.notebook.AddPage(self.rubricPage, "Rubric", True)
        self.notebook.AddPage(self.gradingPage, "Grading", False)
        
        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.Add(self.notebook, 1, wx.EXPAND, 0)
        self.SetSizer(self.mainSizer)
        
        self.mainSizer.SetSizeHints(self)

        # Menu
        self.menu = wx.MenuBar()
        self.SetMenuBar(self.menu)
        fileMenu = wx.Menu()
        fileMenu.Append(ID_NEW, "&New\tCtrl+N", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_SAVE, "&Save\tCtrl+S", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_SAVE_AS, "Save &As", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_LOAD, "&Open\tCtrl+O", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_EXPORT_GRADES, "&Export Grades", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_MERGE, "&Merge", "", wx.ITEM_NORMAL)
        fileMenu.Append(ID_EXIT, "&Quit\tCtrl+Q", "", wx.ITEM_NORMAL)
        self.menu.Append(fileMenu, "&File")

        self.saveMenuID = fileMenu.FindItem("&Save\tCtrl+S")
        self.fileMenu = fileMenu

        actionMenu = wx.Menu()
        actionMenu.Append(ID_RUN, "&Run All Tests\tCtrl+R", "", wx.ITEM_NORMAL)
        actionMenu.Append(ID_RUN_IN_ALL, "Replicate command",
                          "Run a shell command in every submission directory.",
                          wx.ITEM_NORMAL)
        actionMenu.Append(ID_ADD_TO_ALL, "Add file to all",
                          "Copy and add a file to all submissions")
        actionMenu.Append(ID_CONVERT, "Convert for Annotation",
                          "Convert submission files for annotation",
                          wx.ITEM_NORMAL)
        actionMenu.Append(ID_SEND_MAILBACKS, "&Send Mailbacks", "",
                          wx.ITEM_NORMAL)
        actionMenu.Append(ID_WRITE_MAILBACKS, "&Write Mailbacks",
                          "Store to mailback files, do not send",
                          wx.ITEM_NORMAL)
        self.menu.Append(actionMenu, "&Actions")

        optionsMenu = wx.Menu()
        optionsMenu.Append(ID_PROJ_NAME, "Set Project Name")
        optionsMenu.Append(ID_MAIL_OPTIONS, "Set Mail Options")

        self.menu.Append(optionsMenu, "&Options")
                        
    def doLayout(self):
        self.Layout()

    ########################### HANDLERS ###############################
    def onNew(self, event=None):
        """Create a new Agar workspace"""
        if type(event) == type(""):
            filemask = event
            
        dlg = wx.DirDialog(self, "Locate assignment directory",
                           defaultPath=os.getcwd(),
                           style=wx.DEFAULT_DIALOG_STYLE)
        ret = dlg.ShowModal()
        val = str(dlg.GetPath())
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        try:
            os.chdir(os.path.dirname(val))
        except:
            dlg = wx.MessageDialog(self, "Cannot open %s",
                                   "Error",
                                   style=wx.ICON_ERROR|wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            return

        self.a = Agar(os.path.basename(val) + "/", filemask)

        # If we have a self.rubric, then this is being called via the
        # file menu.  Otherwise this was called at startup.  Order
        # matters here.
        if hasattr(self, "rubric"):
            self.rubric.a = self.a
            self.subTree.a = self.a
            self.commentTree.a = self.a
            self.resultTree.a = self.a

            self.rubric.children = []
            self.rubric.update()

            self.subTree.children = []
            self.subTree.update()

            self.commentTree.children = []
            self.commentTree.update()

            self.resultTree.children = []
            self.resultTree.update()
        
        self.filename = None

    def onSave(self, event):
        """Save the file, using the previous filename if there is one,
        and prompting for a new one if there isn't."""
        if self.filename is None:
            dlg = wx.FileDialog(self,
                                style=wx.SAVE|wx.OVERWRITE_PROMPT,
                                wildcard="Agar save files (*.agar)|*.agar")
            ret = dlg.ShowModal()
            fn = dlg.GetPath()
            dlg.Destroy()

            if ret == wx.ID_CANCEL:
                return

            self.filename = fn
            
        self.a.saveAs(self.filename)

    def onSaveAs(self, event):
        """Save the file using a new filename."""
        dlg = wx.FileDialog(self,
                            style=wx.SAVE|wx.OVERWRITE_PROMPT,
                            wildcard="Agar save files (*.agar)|*.agar")
        ret = dlg.ShowModal()
        fn = dlg.GetPath()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        self.filename = fn
        
        self.a.saveAs(fn)

    def onLoad(self, event):
        """Load a saved Agar workspace."""
        dlg = wx.FileDialog(self,
                            style=wx.OPEN|wx.FILE_MUST_EXIST,
                            wildcard="Agar save files (*.agar)|*.agar")
        ret = dlg.ShowModal()
        fn = str(dlg.GetPath())
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        self.load(fn)
        
    def load(self, fn):
        """Actually do the loading.  A little bit of conditional here
        to handle when we are loading at startup vs. loading after
        having started up."""
        
        self.filename = fn
        if hasattr(self, "a"):
            self.a.load(fn)
        else:
            self.a = Agar(".")
            self.a.load(fn)

        if hasattr(self, "commentTree"):
            self.commentTree.update()
            self.subTree.update()
            self.resultTree.update()
            self.rubric.update()

        self.SetTitle("Agar - " + self.a.getAssignmentName())

    def onExport(self, event):
        """Export grades to a spreadsheet."""
        dlg = wx.FileDialog(self,
                            style=wx.OPEN,
                            wildcard="Spreadsheet files (*.gnumeric;*.xls)|"+\
                                     "*.gnumeric;*.xls")
        ret = dlg.ShowModal()
        fn = str(dlg.GetPath())
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        g = GnumeriPy(fn)

        merge = False
        overwrite = True
        if self.a.assignmentName in g.sheetNames():
            dlg = wx.SingleChoiceDialog(self,
                                        "Sheet %s exists." % \
                                        self.a.assignmentName,
                                        "Export Options",
                                        choices=["Overwrite, erasing content",
                                                 "Merge into, preferring new",
                                                 "Merge into, preferring old"])
            ret = dlg.ShowModal()
            val = dlg.GetSelection()
            dlg.Destroy()

            if ret == wx.ID_CANCEL:
                del g
                return

            if val == 0:
                merge = False
                overwrite = True
            if val == 1:
                merge = True
                overwrite = True
            if val == 2:
                merge = True
                overwrite = False

        dlg = wx.MessageDialog(self,
                               "Add column to summary sheet?",
                               "Export options",
                               style=wx.YES_NO|wx.ICON_QUESTION)
        ret = dlg.ShowModal()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            del g
            return

        summary = (ret == wx.ID_YES)

        self.a.exportGnumericScores(fn, summary, overwrite, merge, g)

        del g

    def onMerge(self, event):
        """Merge with another workspace."""

        # Find the file
        dlg = wx.FileDialog(self,
                            style=wx.OPEN|wx.FILE_MUST_EXIST,
                            wildcard="Agar save files (*.agar)|*.agar")
        ret = dlg.ShowModal()
        fn = str(dlg.GetPath())
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return        

        # Prompt for which pieces the user wants to merge (subs,
        # rubric, ucs)
        dlg = MergeDialog(self)
        ret = dlg.ShowModal()
        (r, c) = dlg.GetVals()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return
        
        # Do the merge
        try:
            self.a.merge(fn, r, c)
        except:
            dlg = wx.MessageDialog(self,
                                   "Error in merge: these workspaces " +\
                                   "are incompatible.",
                                   "Error",
                                   style=wx.ICON_ERROR|wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            return

        self.rubric.update()
        self.subTree.update()
        self.commentTree.update()
        self.resultTree.update()
        
    def onExit(self, event):
        """Quit the program, prompting if anything has changed since
        the last save (or if there was no previous save.)"""
        prompt = False
        if not self.filename:
            prompt = True
        else:
            data = [x.replace("\n", "") \
                    for x in gzip.open(self.filename).readlines()]
            if data != self.a.toXML(True, True, True):
                prompt = True

        if prompt:
            dlg = wx.MessageDialog(self,
                                   "Save before quitting?",
                                   "Save?",
                                   style=wx.YES_NO|wx.ICON_QUESTION)
            ret = dlg.ShowModal()
            dlg.Destroy()

            if ret == wx.ID_YES:
                self.onSave(None)                

        sys.exit(0)

    def runCallback(self, username, cur, maxVal):
        """This is what is called by the various test-running
        functions when a new test is being run.  This updates the
        progress dialog and gives the opportunity to cancel the test."""
        if self.progDlg is None:
            self.progDlg = wx.ProgressDialog("Running Tests",
                                             "Testing " + username,
                                             maxVal,
                                             style=wx.PD_APP_MODAL|\
                                             wx.PD_AUTO_HIDE|\
                                             wx.PD_REMAINING_TIME|\
                                             wx.PD_CAN_ABORT)

        if not self.progDlg.Update(cur, "Testing " + username):
            self.progDlg.Destroy()
            self.progDlg = None
            raise KeyboardInterrupt

        if cur == maxVal:
            self.progDlg.Destroy()
            self.progDlg = None

    def onRun(self, event):
        """Run all tests"""        
        self.progDlg = None
        try:
            self.a.runAllTests(self.runCallback)
        except KeyboardInterrupt:
            pass

        self.subTree.update()
        self.resultTree.update()

    def onRunInAll(self, event):
        """Gather a new shell command, run it in each submission directory."""
        dlg = wx.TextEntryDialog(self,
                                 "Enter a command to be executed in each " +\
                                 "submission directory: ",
                                 "Run in all subs")
        ret = dlg.ShowModal()
        val = dlg.GetValue()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        for s in self.a.submissions:
            sub = self.a.submissions[s]
            commands.getoutput("cd %s ; %s " % (sub.getPath(), val))

    def onAddToAll(self, event):
        """Copy and add a file to each submission directory."""
        dlg = wx.FileDialog(self,
                            "Select a file to add to all subs",
                            style=wx.OPEN|wx.FILE_MUST_EXIST)
        ret = dlg.ShowModal()
        val = dlg.GetPath()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        for s in self.a.submissions:
            sub = self.a.submissions[s]

            commands.getoutput("cp %s %s" % (val, sub.getPath()))
            sub.files.append(sub.getPath() + "/" + os.path.basename(val))

        self.subTree.update()

    def convertCallback(self, username, cur, maxVal):
        """Like the run callback, but used for converting to annotatable"""
        if self.progDlg is None:
            self.progDlg = wx.ProgressDialog("Converting Files",
                                             "Converting " + username,
                                             maxVal,
                                             style=wx.PD_APP_MODAL|\
                                             wx.PD_AUTO_HIDE|\
                                             wx.PD_REMAINING_TIME|\
                                             wx.PD_CAN_ABORT)

        if not self.progDlg.Update(cur, "Converting " + username):
            self.progDlg.Destroy()
            self.progDlg = None
            raise KeyboardInterrupt

        if cur == maxVal:
            self.progDlg.Destroy()
            self.progDlg = None

    def onConvertAll(self, event):
        """Do the conversion for annotation."""
        self.a.convertForAnnotation(self.convertCallback)

    def mailCallback(self, text, cur, maxVal):
        """Like the run callback, but used for generating/sending mailbacks"""
        if self.progDlg is None:
            self.progDlg = wx.ProgressDialog("Mailbacks",
                                             text,
                                             maxVal,
                                             style=wx.PD_APP_MODAL|\
                                             wx.PD_AUTO_HIDE|\
                                             wx.PD_REMAINING_TIME|\
                                             wx.PD_CAN_ABORT)

        if not self.progDlg.Update(cur, text):
            self.progDlg.Destroy()
            self.progDlg = None
            raise KeyboardInterrupt

        if cur == maxVal:
            self.progDlg.Destroy()
            self.progDlg = None

    def onSendMailbacks(self, event):
        self.a.sendMailbacks(callback=self.mailCallback)

    def onWriteMailbacks(self, event):
        self.a.genMailbacks(callback=self.mailCallback)

    def onSetProjName(self, event):
        dlg = wx.TextEntryDialog(self,
                                 "Project Name: ",
                                 "Project Name",
                                 defaultValue = self.a.assignmentName)
        ret = dlg.ShowModal()
        val = str(dlg.GetValue())
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        try:
            self.a.setAssignmentName(val)
        except AssertionError:
            dlg = wx.MessageDialog(self,
                                   "Invalid characters in name.",
                                   "Error",
                                   style=wx.ICON_ERROR|wx.OK)
            dlg.ShowModal()
            dlg.Destroy()

        self.SetTitle("Agar - " + self.a.getAssignmentName())

    def onSetMailOptions(self, event):
        """Update the email options"""
        dlg = MailOptionsDialog(self, self.a.reply, self.a.cclist,
                                self.a.domain)
        ret = dlg.ShowModal()
        reply = dlg.reply.GetValue()
        cclist = dlg.cc.GetValue()
        domain = dlg.domain.GetValue()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        self.a.reply = reply
        self.a.cclist = cclist
        self.a.domain = domain
        
    def onSize(self, event):
        """Used to make sure everything resizes correctly."""
        self.mainSizer.SetSizeHints(self)
        self.Layout()
    
class AgarApp(wx.App):
    def OnInit(self):
        frame = AgarFrame(None, -1, "")
        self.SetTopWindow(frame)
        frame.Show(1)
        return 1

if __name__ == "__main__":
    agar = AgarApp()
    agar.MainLoop()
    del agar

    


