# Module stolen from Idle's PyParse module.
# It reads a module, and builds a list of entries that describe each line.
# The description entries are:
# <blank> blank line or comment
# 's' - single line statement
# 'f' - first line of multi-line statement
# 'm' - middle line of multi-line statement
# 'l' - last line of multi-line statement
# 'x' - "else:" or "except" statement. These get skipped by the tracer.

import string # XXX see if we can eliminate this.
import re
import sys
import getopt
import pickle
import os


# Reason last stmt is continued (or C_NONE if it's not).
C_NONE, C_BACKSLASH, C_STRING, C_BRACKET = range(4)

if 0:   # for throwaway debugging output
    def dump(*stuff):
        sys.__stdout__.write(string.join(map(str, stuff), " ") + "\n")

# Build translation table to map uninteresting chars to "x", open
# brackets to "(", and close brackets to ")".

_tran = ['x'] * 256
for ch in "({[":
    _tran[ord(ch)] = '('
for ch in ")}]":
    _tran[ord(ch)] = ')'
for ch in "\"'\\\n#":
    _tran[ord(ch)] = ch
_tran = string.join(_tran, '')
del ch

try:
    UnicodeType = type(unicode(""))
except NameError:
    UnicodeType = None

# As quickly as humanly possible <wink>, find the line numbers (0-
# based) of the non-continuation lines.
# Creates self.{goodlines, continuation}.

class Parser:
    def __init__(self):
        self.lineStatus = 's'
    
    def setContinue(self, isLast):
        if isLast and self.lineStatus == 's':
            return 's'
        if isLast and self.lineStatus != 's':
            self.lineStatus = 's'
            return 'l'
        if self.lineStatus == 's':
            self.lineStatus = 'c'
            return 'f'
        assert self.lineStatus == 'c'
        return self.lineStatus

    def _study1(self, inList, _replace=string.replace, _find=string.find):

        self.str = "".join(inList)
        outList = ['?'] * len(inList)

        # Map all uninteresting characters to "x", all open brackets
        # to "(", all close brackets to ")", then collapse runs of
        # uninteresting characters.  This can cut the number of chars
        # by a factor of 10-40, and so greatly speed the following loop.
        str = self.str
        str = string.translate(str, _tran)
        str = _replace(str, 'xxxxxxxx', 'x')
        str = _replace(str, 'xxxx', 'x')
        str = _replace(str, 'xx', 'x')
        str = _replace(str, 'xx', 'x')
        str = _replace(str, '\nx', '\n')
        # note that replacing x\n with \n would be incorrect, because
        # x may be preceded by a backslash

        # March over the squashed version of the program, accumulating
        # the line numbers of non-continued stmts, and determining
        # whether & why the last stmt is a continuation.
        continuation = C_NONE
        level = lno = 0     # level is nesting level; lno is line number
        self.goodlines = goodlines = [0]
        push_good = goodlines.append
        i, n = 0, len(str)
        while i < n:
            ch = str[i]
            i += 1

            # cases are checked in decreasing order of frequency
            if ch == 'x':
                continue

            if ch == '\n':
                outList[lno] = self.setContinue(level == 0)
                lno += 1
                continue

            if ch == '(':
                level = level + 1
                continue

            if ch == ')':
                if level:
                    level = level - 1
                    # else the program is invalid, but we can't complain
                continue

            if ch == '"' or ch == "'":
                # consume the string
                quote = ch
                if str[i-1:i+2] == quote * 3:
                    quote = quote * 3
                w = len(quote) - 1
                i = i+w
                while i < n:
                    ch = str[i]
                    i = i+1

                    if ch == 'x':
                        continue

                    if str[i-1:i+w] == quote:
                        i = i+w
                        break

                    if ch == '\n':
                        lno += 1
                        if w == 0:
                            # unterminated single-quoted string
                            outList[lno-1] = self.setContinue(level == 0)
                            break
                        outList[lno-1] = self.setContinue(0) # not-last line of multi-line comment
                        continue

                    if ch == '\\':
                        assert i < n
                        if str[i] == '\n':
                            lno = lno + 1
                        i = i+1
                        continue

                    # else comment char or paren inside string

                else:
                    # didn't break out of the loop, so we're still
                    # inside a string
                    continuation = C_STRING
                continue    # with outer loop

            if ch == '#':
                # consume the comment
                i = _find(str, '\n', i)
                assert i >= 0
                continue

            assert ch == '\\'
            assert i < n
            if str[i] == '\n':
                outList[lno] = self.setContinue(0)
                lno += 1
                if i+1 == n:
                    continuation = C_BACKSLASH
            i += 1

        # The last stmt may be continued for all 3 reasons.
        # String continuation takes precedence over bracket
        # continuation, which beats backslash continuation.
        if continuation != C_STRING and level > 0:
            continuation = C_BRACKET
        self.continuation = continuation

        # the last line does not seem to be evaluated. If it's still a '?',
        # we'll plug it with either an 'l' or an 's'.

        if outList[-1] == "?":
            if outList[-2] == "s":
                outList[-1] = 's'
            else:
                outList[-1] = 'l'


        # new code starts here ---
        i = 0
        n = len(inList)
        # create array with first lines indicated
        while i < n:
            if outList[i] == 's':
                testLine = inList[i].strip()
                if len(testLine) == 0 or testLine[0] == "#":
                    outList[i] = " "
                elif testLine[:5] == "else:" or testLine[:6] == "except":
                    outList[i] = "x"
            i += 1
        return outList
    
# Following code originally from pyviewc 0.2

def processFile(outList, lineNos):
    i = 0
    n = len(outList)
    outList2 = ["?"] * n
    numUncovered = 0


    while i < n:
        if outList[i] == ' ':
            outList2[i] = ' '
        else:
            outList2[i] = '!'
        i += 1
    for lineNo in lineNos:
        outList2[lineNo-1] = "1"
        
    i = 0
    while i < n:
        if outList[i] in ('c', 'l') and outList2[i-1] in ('1', '2', '3'):
            outList2[i] = '2'
        elif outList[i] == 'x' and outList2[i+1] in ('1', '2', '3'):
            outList2[i] = '3'
        i += 1

    i = n - 2
    while i >= 0:
        if outList2[i] == '!':
            numUncovered += 1
            if outList[i] in ('c', 'f') and outList2[i+1] in ('1', '2', '3', '4'):
                outList2[i] = '4'
                numUncovered -= 1
        i -= 1

    return (outList2, numUncovered)


def getCoverInfo(filename):
    f=open(filename, 'r')
    dict = pickle.load(f)

    k=dict.keys()
    k.sort()

    #f = open('pyviewc.pickle.wri', 'w')
    #for file, lineNo, event in k:
    #    f.write('%s %s %s\n' % (file, lineNo, event))
    #f.close()

    fileDict = {}
    for file, lineNo, event in k:
        if not fileDict.has_key(file):
            fileDict[file] = list((lineNo,))
        else:
            fileDict[file].append(lineNo)
    f.close()
    return fileDict

def exclude(file):
    if file[:len('/usr/local/')]=='/usr/local/': return 1
    if file[0]=='<': return 1
    if os.path.basename(file)=='pycoverfuncs.py': return 1
    if file[:16].lower() == "c:\\python22\\lib\\": return 1
    if file[:30].lower() == "c:\\python22\\mystuff\\testtools\\": return 1
    return 0

def printUsage():
    print """pyviewc: coverage data viewer

usage:
    pyviewc [options] [files]
    
options:
    -ix, --input=x    which file to read coverage data from
    -h, --help        this screen
    -ere, --exclude=re  exclude files that match the regular expression re
    -Ire, --include=re  include files that match the regular expression re

    * --include's override --excludes
    
    files: which files to output coverage data for
"""
    sys.exit(0)

def getOptions():
    filename=None
    outputs=[]
    excludere=None
    includere=None
    
    try:
        ret=getopt.getopt(sys.argv[1:], "i:e:I:h", ["input=",
                                                    'exclude=',
                                                    'include=',
                                                    'help'])
    except:
        printUsage()

    for i in ret[0]:
        if i[0]=='--input' or i[0]=='-i':
            filename=i[1]
        elif i[0]=='--help' or i[0]=='-h':
            printUsage()
        elif i[0]=='-e' or i[0]=='--exclude':
            excludere = re.compile(i[1])
        elif i[0]=='-I' or i[0]=='--include':
            includere = re.compile(i[1])

    return (filename, ret[1], excludere, includere)

def lookingFor(filename, files):
    if not files: 
        return 1
    for i in files:
        if os.path.basename(i)==os.path.basename(filename):
            return 1
    return 0

if __name__=='__main__':
    
    (coverFile, files, excludere, includere)=getOptions()

    if coverFile is None:
        try:
            coverFile=os.environ['PYCOVERFILE']
        except:
            coverFile='coverage.dat'

    fileDict=getCoverInfo(coverFile)

    for i in fileDict.keys():
        if not lookingFor(i, files):
            print "File '%s' examined. Excluded." % (i,)
            # by lookingFor test" % (i,)
            continue
            
        if exclude(i) or (excludere and excludere.search(i)):
            if not (includere and includere.search(i)):
                print "File \"%s\" examined. Excluded. " % (i,)
                continue
        try:
            inFile=open(i,'r')
        except:
            print 'unable to open %s!, skipping...' % i
            continue
        inList = inFile.readlines()
        inFile.close()

        try:
            outFileName = i[string.rindex(i,'/')+1:]
            print 'using stripped name', outFileName
        except:
            outFileName = i
        outFileName = outFileName + '.cov'

        parser = Parser()
        outList = parser._study1(inList)
        outList2, numUncovered = processFile(outList, fileDict[i])
        print "File \"%s\" processed. Number of uncovered lines: %s" % \
              (i, numUncovered)
        if numUncovered != 0:
            try:
                outFile=open(outFileName,'w')
                #outFile.writelines(map(lambda x, y, z: x + " " + y + " " + z,
                #                       outList, outList2, inList))
                outFile.writelines(map(lambda x, y, z: ((y in ' !') and y or '>') + " " + z,
                                       outList, outList2, inList))
                outFile.close()
            except:
                print "couldn't open %s! skipping..." % outFileName
                continue

