import fnmatch
import commands
import string

from xml.sax.saxutils import escape, quoteattr

import tests
import comments

class RubricTest:
    """

    Major Members
    * self.name -- the name we are identified with in the Rubric tree
    * self.test -- which Test are we using?
    * self.rubric -- the Rubric we are part of (there should be only one)
    * 
    * self.fileMask -- a comma-separated list of globs describing which
    files this test operates on.
    * self.points -- how many points is this worth?
    """
    def __init__(self, rubric, test, points=1.0, name=None):
        # Get a valid name
        if name == None:
            name = test.name
            
        # Make sure we aren't duplicating anything
        if rubric and rubric.hasItem(name):
            name = rubric.nextName(name)

        # Do the update
        self.name = name
        if rubric:
            rubric.addMapping(self)
        
        # Remember our parent
        self.rubric = rubric
            
        # Which test are we?
        self.test = test

        # How many points is this worth?
        self.points = points

        # We get our default filemask from the master test
        if type(self.test) != type(""):
            self.fileMask = self.test.fileMask

        # Do we have any children?
        self.children = []

        # Auto-comments on success and failure
        self.successComments = []
        self.failureComments = []

        # Configuration options (command line parameters)
        if type(self.test) != type(""):
            self.config = {}
            for key in self.test.defaultConfig:
                self.config[key] = self.test.defaultConfig[key]

    def toXML(self):
        ret = []
        ret.append('<rubricItem name=%s testName=%s fileMask=%s points="%f">'%\
                   (quoteattr(self.name),
                    quoteattr(self.test.name),
                    quoteattr(self.fileMask),
                    self.points))

        # Children
        ret.append("<children>")
        for c in self.children:
            ret.append('<child>%s</child>" />' % escape(c.name))
        ret.append("</children>")
        
        # Auto-comments
        ret.append('<autocomments success="1">')
        for c in self.successComments:
            ret += c.toXML()
        ret.append('</autocomments>')
        ret.append('<autocomments success="0">')
        for c in self.failureComments:
            ret += c.toXML()
        ret.append('</autocomments>')

        # Config
        ret.append("<config>")
        for key, val in self.config.items():
            ret.append('<configitem key="%s" value="%s" />' % (key, val))
        ret.append("</config>")

        ret.append('</rubricItem>')

        return ret

    def hasChild(self, test):
        """Is test the same as this one or a descendent?"""
        if test == self: return True

        for child in self.children:
            if child.hasChild(test): return True

        return False

    def newChild(self, test, points=1.0, name=None):
        """Add a child to this RubricTest (another RubricTest), and
        return it."""
        newTest = RubricTest(self.rubric, test, points, name)
        self.children.append(newTest)
        return newTest

    def removeChild(self, test, complete=True):
        """Remove the given test from our children list.  If the
        optional second parameter is not False, also remove the given
        test from the rubric mapping (calling delMapping)."""
        if test in self.children:
            self.children.remove(test)

        if complete:
            self.rubric.delMapping(test)

    def getName(self):
        """Return the name of this test"""
        return self.name

    def setName(self, name):
        """Change the name of this RubricTest"""
        if self.rubric.hasItem(name):
            raise ValueError
        self.rubric.delMapping(self)
        self.name = name
        self.rubric.addMapping(self)
    
    def setFileMask(self, fileMask):
        """Overwrite the fileMask member."""
        assert(type(fileMask) == type(""))
        self.fileMask = fileMask

    def getFileMask(self):
        """Get the fileMask member. """
        return self.fileMask

    def setPoints(self, points):
        """Overwrite the points member. """
        assert(type(points) in [type(0), type(0.0)])
        self.points = points

    def getPoints(self):
        """Return the points member. """
        return self.points

    def addAutoComment(self, onFail, text, value, op, appliesToAll=False):
        """If onFail is True, then a new comment (in addition to the
        test-result comment) is assigned to the submission whenever
        this test fails.  Otherwise, the given comment is assigned
        when this test succeeds.
        """
        assert(onFail in [True, False])
        comment = comments.Comment(self, True, value, text, op, appliesToAll)
        
        if onFail:
            self.failureComments.append(comment)
        else:
            self.successComments.append(comment)

    def removeAutoComment(self, onFail, text, value, op, appliesToAll=False):
        """Remove the comment with the given comment text, value, and
        op from the appropriate auto-comment list, if it exists."""
        assert(onFail in [True, False])

        # Should remove all copies of this one from submissions.
        tempComment = comments.Comment(self, True, value,
                                       text, op, appliesToAll)
        self.rubric.agar.removeAllCopiesOfComment(tempComment)
        
        if onFail:
            for comment in self.failureComments:
                if comment.comment == text and \
                   comment.value == value and \
                   comment.op == op:
                    self.failureComments.remove(comment)
                    break
        else:
            for comment in self.successComments:
                if comment.comment == text and \
                   comment.value == value and \
                   comment.op == op:
                    self.successComments.remove(comment)
                    break
        

    def recursiveRemoveAutoComments(self, submission):
        """Make sure that nothing in this branch of the tree is
        leaving false auto-comments in this submission."""        
        submission.removeAutoComments(self)
        for child in self.children:
            child.recursiveRemoveAutoComments(submission)
            
    def runTest(self, submission):
        """Run this test on this submission. Returns a Comment."""

        # If we are a top-level, make sure none of our chilren have
        # remaining auto-comments
        if self.rubric.isTopLevel(self):
            self.recursiveRemoveAutoComments(submission)
        
        # The unique set of files that match one or more of the
        # comma-separated filemasks provided by this test
        allFiles = []
        for mask in self.fileMask.split(","):
            allFiles += fnmatch.filter(submission.getFiles(), mask.strip())
        files = list(set(allFiles))
        
        results = []

        cmdLines = []
        # Do we feed in files one-at-a-time or all at once?
        if self.test.allAtOnce:
            cmdLines = ["%s %s %f" % (self.getCommandLine(),
                                      " ".join(files),
                                      self.points)]
        else:
            for fileName in files:
                cmdLines.append("%s %s %f" % (self.getCommandLine(),
                                              fileName,
                                              self.points))

        # If we got nothing, then 
        if not cmdLines:
            return [comments.Comment(self, True, 0,
                                     "No files matching " + self.fileMask,
                                     "+")]
        

        # Run those command lines
        for cmdLine in cmdLines:
            status, output = self.test.run(cmdLine, submission)

            output = ascii(output)

            if status:
                return [comments.Comment(self, True, 0,
                                         "Unexpected error in test:\n"+ \
                                         output, "+")]
            results.append(output)

        # Extract point values and textual output 
        vals = [(float(x.split("\n")[0]), x) for x in results]

        choice = min(vals)

        points = choice[0]
        assert(type(points) == type(0.0))
        output = "\n".join(choice[1].split("\n")[1:])

        # These are the comments we are going to apply, so far
        ret = [comments.Comment(self, True, points, output, "+")]

        # If we've got auto-comments to apply, apply 'em.
        if self.successComments and points:
            ret += self.successComments
        if self.failureComments and not points:
            ret += self.failureComments

        # If we failed and have children, run the children
        if self.children and not points:
            for child in self.children:
                ret += child.runTest(submission)
        
        return ret

    def setOption(self, name, val):
        """Add a parameter to our configuration for this test"""
        assert(name in self.test.getParams())
        self.config[name] = str(val)

    def unsetOption(self, name):
        """Remove a parameter from our configuration for this test"""
        assert(name in self.config)
        del self.config[name]
        
    def getCommandLine(self):
        """Return a reasonable command line given our configuration
        (self.config)"""
        ret = ""
        for op in self.test.getFlagParams():
            if op in self.config:
                ret += "--%s " % op.replace(" ", "-")

        for op in self.test.getValParams() + self.test.getFileParams():
            if op in self.config:
                ret += "--%s='%s' " % (op.replace(" ", "-"),
                                       self.config[op].replace(" ", "_"))

        return ret
    
class Rubric:
    """The main Rubric class.  Keeps track of the toplevel RubricTests
    (those that are always run), as well as the complete name-to-test
    mapping for the whole Rubric tree.

    Major Members:
    * self.topLevel -- a list of RubricTests, the 'top level' rubric items
    * self.rubricItems -- a dictionary mapping names to RubricTests
    """
    def __init__(self, agar):
        self.topLevel = []
        self.rubricItems = {}
        self.agar = agar

    def set(self, newRubric, order):
        self.topLevel[:] = []
        self.rubricItems = newRubric

        for name in order:
            self.topLevel.append([x for x in newRubric.values() \
                                  if x.name == name][0])

        for val in self.rubricItems.values():
            val.test = self.agar.tests[val.test]
            val.rubric = self

    def merge(self, newRubric, order, preferOld):
        oldNames = [x.name for x in self.topLevel]

        for name in order:
            if name in oldNames:
                if preferOld:
                    # This will only return one thing, which is the test that
                    # is overlapping
                    overlap = [x for x in self.topLevel \
                               if x.name == name][0]
                    ind = self.topLevel.index(overlap)

                    self.rubric.topLevel[ind] = newRubric[name]

            else:
                self.topLevel.append(newRubric[name])
                self.addMapping(self.topLevel[-1])
                    
    def __getitem__(self, key):
        return self.rubricItems[key]

    def toXML(self):
        ret = []
        ret.append("<rubric>")
        
        ret.append("<rubricTests>")
        testNames = self.rubricItems.keys()
        testNames.sort()
        for testName in testNames:
            ret += self.rubricItems[testName].toXML()
        ret.append("</rubricTests>")

        ret.append("<topLevel>")
        for test in self.topLevel:
            ret.append('<testname>%s</testname>' % escape(test.name))
        ret.append("</topLevel>")

        ret.append("</rubric>")
        return ret

    def newTopLevel(self, test, points=1.0, name=None):
        """Get a new top-level rubric item"""
        newTest = RubricTest(self, test, points, name)
        self.addToTopLevel(newTest)
        return newTest

    def hasItem(self, name):
        """Is this name in use?"""
        return name in self.rubricItems

    def nextName(self, name):
        """Since we want unique name-to-RubricTest mapping, we can't
        have duplicate names.  If a name would be duplicated, this
        finds the next name of the form 'name + #num' that is not in
        use, and returns that.  It does _not_ insert the test into the
        rubric.
        """
        if name not in self.rubricItems:
            return name

        n = 1
        while name + " #%d" % n in self.rubricItems:
            n += 1

        return name + " #%d" % n

    def addToTopLevel(self, rubricTest,n=None):
        """Add to the top-level of the rubric"""
        if n == None:
            n = len(self.topLevel)
            
        self.topLevel.insert(n, rubricTest)

    def removeFromTopLevel(self, rubricTest):
        """Remove from the top-level of the rubric, if present"""
        if rubricTest in self.topLevel:
            self.topLevel.remove(rubricTest)

    def addMapping(self, rubricTest):
        """Add to the listing of RubricTests"""
        assert(rubricTest.name not in self.rubricItems)
        self.rubricItems[rubricTest.name] = rubricTest

    def delMapping(self, rubricTest):
        """Remove from the listing of RubricTests"""        
        assert(rubricTest.name in self.rubricItems)
        del self.rubricItems[rubricTest.name]
        
    def isTopLevel(self, toFind):
        """Find whether the given rubric test, by name or by value, is
        in the toplevel.  Return bool"""
        
        if type(toFind) == type(""):
            for test in self.topLevel:
                if test.getName() == toFind:
                    return True
            return False

        assert(isinstance(toFind, RubricTest))

        return toFind in self.topLevel

    def getTopLevel(self):
        return self.topLevel[:]

    def getNames(self):
        """Give the names of the toplevel tests, in order."""
        return [test.getName() for test in self.topLevel]


def ascii(s):
    s = str(s)
    ret = ""
    for char in s:
        if ord(char) <= 127:
            ret += char

    return ret
