import string

import rubric

from memTree import *
from droptarget import *
from TestConfigDialog import *


class RubricTree(MemTree):
    """A memtree based widget for setting up a rubric of rubrictests"""
    def __init__(self, parent, agar, *args, **kwds):
        MemTree.__init__(self, "Rubric", *args, **kwds)

        self.a = agar
        self.parent = parent
        self.update()

        self.copied = None

        self.dt = AgarDropTarget(self.getDropData)
        self.data = wx.CustomDataObject(wx.CustomDataFormat("AgarTest"))
        self.dt.SetDataObject(self.data)
        
        self.SetDropTarget(self.dt)

        # Configure, delete, copy, paste, rename, run
        self.contextConfigureID = wx.NewId()
        self.contextDeleteID = wx.NewId()
        self.contextCopyID = wx.NewId()
        self.contextPasteID = wx.NewId()
        self.contextRenameID = wx.NewId()
        self.contextRunID = wx.NewId()
        self.contextUpID = wx.NewId()
        self.contextDownID = wx.NewId()
        self.contextRightID = wx.NewId()
        self.contextLeftID = wx.NewId()

        wx.EVT_MENU(self, self.contextConfigureID, self.onConfigContext)
        wx.EVT_MENU(self, self.contextDeleteID, self.onDeleteContext)
        wx.EVT_MENU(self, self.contextCopyID, self.onCopyContext)
        wx.EVT_MENU(self, self.contextPasteID, self.onPasteContext)
        wx.EVT_MENU(self, self.contextRenameID, self.onRenameContext)
        wx.EVT_MENU(self, self.contextRunID, self.onRunContext)
        wx.EVT_MENU(self, self.contextUpID, self.onUpContext)
        wx.EVT_MENU(self, self.contextDownID, self.onDownContext)
        wx.EVT_MENU(self, self.contextLeftID, self.onLeftContext)
        wx.EVT_MENU(self, self.contextRightID, self.onRightContext)
                                
        wx.EVT_RIGHT_UP(self, self.onRightUp)
        wx.EVT_TREE_ITEM_ACTIVATED(self, -1, self.onDClick)

        wx.EVT_TREE_END_LABEL_EDIT(self, -1, self.onEditEnd)
        wx.EVT_TREE_BEGIN_LABEL_EDIT(self, -1, self.onEditBegin)

    def getDropData(self, x, y, data):
        """Allow drag-and-drop"""
        if self.dt.GetData():
            testName = str(self.data.GetData())            

            hitId = self.HitTest((x, y))[0]

            # We were dropped on something:
            atEnd = False
            if hitId.IsOk():
                hitString = str(self.GetItemText(hitId))
                if hitString == "Rubric":
                    atEnd = True
                else:
                    parentName = hitString
            else:
                atEnd = True

            if atEnd:
                if testName not in self.a.tests:
                    return
                
                self.a.rubric.newTopLevel(self.a.tests[testName])
            else:
                newTest = rubric.RubricTest(self.a.rubric,
                                            self.a.tests[testName])
                                            
                self.a.rubric[parentName].newChild(self.a.tests[testName])

        self.update()

    def update(self):
        """Draw the thing"""
        self.unTag()
        for top in self.a.rubric.getTopLevel():
            self.add(top)

        self.prune()
        MemTree.update(self)
        
    def add(self, rubricItem, prev=None):
        """Add to the end, or if prev was specified, add as a child of
        that."""
        if prev is None:
            prev = []

        prev.append(rubricItem.getName())
        self.pathAdd(prev)

        for child in rubricItem.children:
            self.add(child, prev)

    def getRubricTest(self):
        """Find the current rubric test"""
        name = self.GetItemText(self.GetSelection())
        assert(self.GetSelection() != self.tree.id)

        return self.a.rubric[name]

    def onEditEnd(self, event):
        oldName = str(self.GetItemText(event.GetItem()))
        newName = str(event.GetLabel())
        self.rename(oldName, newName, event)

    def onEditBegin(self, event):
        name = str(event.GetLabel())
        if name == "Rubric":
            event.Veto()

    def rename(self, oldName, newName, event):
        """Rename a rubric item."""
        allowedChars = string.ascii_letters + string.digits + " #_"
        good = True
        for char in newName:
            if char not in allowedChars:
                good = False
                break

        if not good:
            dlg = wx.MessageDialog(self,
                                   "Rubric item names can only be made of " + \
                                   "alphanumeric characters, spaces, #, and _",
                                   "Notice",
                                   wx.ICON_ERROR|wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            event.Veto()

        # All is well, do the update
        test = self.a.rubric[oldName]
        try:
            test.setName(newName)
        except ValueError:
            dlg = wx.MessageDialog(self,
                                   "Rubric item names must be unique.",
                                   "Error",
                                   wx.ICON_ERROR|wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
            event.Veto()

        # Find the item in the tree and change the name
        self.renameItem(oldName, newName)

        self.update()
        self.parent.resultTree.update()

    def renameItem(self, oldName, newName, cur=None):
        """Do the rename"""
        if cur is None:
            cur = self.tree

        if cur.name == oldName:
            cur.name = newName
            return

        for child in cur.children:
            self.renameItem(oldName, newName, child)
            
    def onRightUp(self, event):
        """Display the context menu"""
        try:
            test = self.getRubricTest()
        except AssertionError:
            return
        except KeyError:
            return
        
        menu = wx.Menu()
        menu.Append(self.contextConfigureID, "Configure")
        menu.Append(self.contextDeleteID, "Delete")
        menu.Append(self.contextCopyID, "Copy")
        if self.copied:
            menu.Append(self.contextPasteID, "Paste")
        menu.Append(self.contextRenameID, "Rename")
        menu.Append(self.contextRunID, "Run Test")        

        selId = self.GetSelection()
        up = self.GetPrevSibling(selId)
        down = self.GetNextSibling(selId)

        menu.AppendSeparator()

        if up.IsOk():
            menu.Append(self.contextUpID, "Move Up")
        if down.IsOk():
            menu.Append(self.contextDownID, "Move Down")
        if up.IsOk():
            menu.Append(self.contextRightID, "Make sub-test of " + \
                        str(self.GetItemText(up)))
        if self.GetItemParent(selId) != self.GetRootItem():
            menu.Append(self.contextLeftID, "Make sibling of " + \
                        str(self.GetItemText(self.GetItemParent(selId))))

        self.PopupMenu(menu, event.GetPosition())
        menu.Destroy()

    def onDClick(self, event):
        """Run the configuration dialog"""
        self.onConfigContext(event)

    def onConfigContext(self, event):
        """Run the configuration dialog"""
        try:
            test = self.getRubricTest()
        except:
            return
        
        dlg = TestConfigDialog(self, test)
        dlg.ShowModal()
        dlg.Destroy()

    def onDeleteContext(self, event):
        """Delete a rubric item"""
        try:
            test = self.getRubricTest()
        except:
            return

        # If this has user comments, prompt for confirmation before deleting
        # those.
        found = False
        for cname in self.a.userComments:
            c = self.a.userComments[cname]

            if c.test == test:
                found = True
                break

        if found:
            dlg = wx.MessageDialog(self,
                                   "This will delete all associated user " +\
                                   "comments, continue?",
                                   "Confirm",
                                   style=wx.YES_NO|wx.ICON_QUESTION)
            ret = dlg.ShowModal()
            dlg.Destroy()

            if ret == wx.ID_NO:
                return
            
        # If it is toplevel, remove it from there
        if test in self.a.rubric.getTopLevel():
            self.a.rubric.removeFromTopLevel(test)
            self.a.rubric.delMapping(test)
            self.update()
            return

        # Otherwise, find its parent and remove it from there
        parentName = str(self.GetItemText(self.GetItemParent(\
            self.GetSelection())))

        parent = self.a.rubric[parentName]

        parent.removeChild(test)

        # Remove assigned comments
        for s in self.a.submissions:
            sub = self.a.submissions[s]

            ind = 0
            while ind < len(sub.comments):
                if test.hasChild(sub.comments[ind].test):
                    sub.comments.remove(sub.comments[ind])
                else:
                    ind += 1

        # Remove 
        for cname in self.a.userComments.keys():
            c = self.a.userComments[cname]

            # User comments can only be assigned to top-level tests
            if c.test == test:
                del self.a.userComments[c]

        self.update()                                 

    def onCopyContext(self, event):
        """Copy the current rubric item 'to the clipboard'"""
        self.copied = self.getRubricTest()

    def doCopy(self, cur, dest):
        """Paste the copied rubric item"""
        # Make a new one of cur and put it in dest
        newTest = rubric.RubricTest(self.a.rubric, cur.test,
                                    cur.points, cur.name)

        # Copy other settings
        for key in cur.config:
            newTest.config[key] = cur.config[key]

        newTest.successComments = cur.successComments[:]
        newTest.failureComments = cur.failureComments[:]
        newTest.fileMask = cur.fileMask

        # Add it in
        dest.append(newTest)

        # And copy over the children
        for child in cur.children:
            self.doCopy(child, newTest.children)

        # And redraw
        self.update()        

    def onPasteContext(self, event):
        """Paste the copied rubric item"""
        self.doCopy(self.copied, self.a.rubric.topLevel)

    def onRenameContext(self, event):
        """Rename a rubric item"""
        dlg = wx.TextEntryDialog(self, "New name", "New name",
                                 style=wx.OK|wx.CANCEL)
        ret = dlg.ShowModal()
        val = str(dlg.GetValue())
        dlg.Destroy()

        oldName = str(self.GetItemText(self.GetSelection()))
        newName = val

        try:
            self.rename(oldName, newName, None)
        except:
            pass

    def onRunContext(self, event):
        """Run just one test, if we actually right clicked on a test"""
        self.progDlg = None
        try:
            self.a.runTest(self.getRubricTest(), self.parent.runCallback)
        except KeyboardInterrupt:
            return
        except AssertionError:
            return
        except KeyError:
            return

        self.parent.subTree.update()
        self.parent.resultTree.update()

    def getContainingList(self, test, cur=None):
        """Find the list of rubric items the given test is contained
        in, and return (by reference) the containing list"""
        if cur == None:
            cur = self.a.rubric.topLevel

        if test in cur:
            return cur

        for child in cur:
            temp = self.getContainingList(test, child.children)
            if temp is not None:
                return temp

        return None
    
    def onUpContext(self, event):
        """Move this up the tree (swap with previous sibling)"""
        test = self.getRubricTest()
        
        l = self.getContainingList(test)
        assert(l is not None)
        ind = l.index(test)


        # Swap 'em
        l[ind - 1], l[ind] = l[ind], l[ind - 1]

        self.tree.children = []
        self.update()            

    def onDownContext(self, event):
        """Move this down the tree (swap with next sibling)"""
        test = self.getRubricTest()

        l = self.getContainingList(test)
        assert(l is not None)
        ind = l.index(test)

        # Swap 'em
        l[ind + 1], l[ind] = l[ind], l[ind + 1]

        self.tree.children = []
        self.update()            


    def onRightContext(self, event):
        """Indent this into the tree (make sub-test of previous sibling)"""
        test = self.getRubricTest()

        # Confirm and tweak all user comments assigned to this test
        # (both stored and assigned)
        found = False
        for cname in self.a.userComments:
            c = self.a.userComments[cname]

            if c.test == test:
                found = True

        if found:
            dlg = wx.MessageDialog(self,
                                   "This will reassign all user comments " + \
                                   "assigned to this test to the new parent "+\
                                   "test.  Continue?",
                                   "Warning",
                                   style=wx.YES_NO|wx.ICON_INFORMATION)
            ret = dlg.ShowModal()
            dlg.Destroy()
            if ret == wx.ID_NO:
                return

        l = self.getContainingList(test)
        assert(l is not None)
        ind = l.index(test)

        newParent = l[ind - 1]
        newParent.children.append(test)
        l.remove(test)

        # Do the updates, since this is all by-reference this should suffice.
        for cname in self.a.userComments:
            c = self.a.userComments[cname]

            if c.test == test:
                c.test = newParent
                

        self.tree.children = []
        self.update()
        self.parent.subTree.update()
        self.parent.resultTree.update()

    def onLeftContext(self, event):
        """Make this the next-sibling of the item that is currently
        its parent."""
        test = self.getRubricTest()

        parentName = str(self.GetItemText(
            self.GetItemParent(self.GetSelection())))
        parent = self.a.rubric[parentName]

        l = self.getContainingList(parent)

        parent.children.remove(test)
        ind = l.index(parent)
        l.insert(ind, test)

        self.tree.children = []
        self.update()

    
