import os
import commands

from memTree import *
from CommentDialog import *
from AnnotateFrame import *

class SubTree(MemTree):
    """A widget that holds all the submissions, their alternate
    usernames, the files for each submission, and the comments for
    each submission."""
    def __init__(self, parent, agar, *args, **kwds):
        MemTree.__init__(self, "Submissions", *args, **kwds)

        # This lists the PIDs of external viewers used to view files
        # and such.  We should call waitpid on them occasionally.  (We
        # do so in update())
        self.pids = []

        self.parent = parent
        self.a = agar
        self.update()

        # Menu IDs
        self.contextNewSubID = wx.NewId()
        self.contextDelSubID = wx.NewId()
        self.contextNewFileID = wx.NewId()
        self.contextDelFileID = wx.NewId()
        self.contextOpenFileID = wx.NewId()
        self.contextRunTestsID = wx.NewId()
        self.contextDelCommentID = wx.NewId()
        self.contextEditCommentID = wx.NewId()
        self.contextNewUsernameID = wx.NewId()
        self.contextDelUsernameID = wx.NewId()

        wx.EVT_MENU(self, self.contextNewSubID, self.onNewSub)
        wx.EVT_MENU(self, self.contextDelSubID, self.onDelSub)
        wx.EVT_MENU(self, self.contextNewFileID, self.onNewFile)
        wx.EVT_MENU(self, self.contextDelFileID, self.onDelFile)
        wx.EVT_MENU(self, self.contextOpenFileID, self.onOpenFile)
        wx.EVT_MENU(self, self.contextRunTestsID, self.onRunTests)
        wx.EVT_MENU(self, self.contextDelCommentID, self.onDelComment)
        wx.EVT_MENU(self, self.contextEditCommentID, self.onEditComment)
        wx.EVT_MENU(self, self.contextNewUsernameID, self.onNewUsername)
        wx.EVT_MENU(self, self.contextDelUsernameID, self.onDelUsername)

        wx.EVT_RIGHT_UP(self, self.onRightUp)

        wx.EVT_TREE_ITEM_ACTIVATED(self, -1, self.onDClick)

        wx.EVT_KEY_DOWN(self, self.onKeyDown)

        self.findViewers()

    def findViewers(self):
        """Find the basic viewers for the common file types."""
        self.viewers = {}

        toTest = [("acroread", "pdf"),
                  ("xpdf", "pdf"),
                  ("ggv", "ps"),
                  ("gv", "ps"),
                  ("abiword", "doc", "rtf"),
                  ("oowriter", "doc", "rtf"),
                  ("eog", "gif", "jpg", "jpeg", "png"),
                  ("xv", "gif", "jpg", "jpeg", "png"),
                  ("emacs", "c", "cc", "cpp", "h", "txt", "makefile","readme")]
                  
        for i in toTest:
            if not commands.getstatusoutput("which " + i[0])[0]:
                for j in i[1:]:
                    self.viewers[j] = self.viewers.get(j, []) + [i[0]]

        self.defaultViewer = "emacs"

    def getUsername(self):
        """Get the username of the current submission"""
        curId = self.GetSelection()
        root = self.GetRootItem()

        assert(curId != root)

        while self.GetItemParent(curId) != root:
            curId = self.GetItemParent(curId)

        return self.GetItemText(curId)

    def update(self):
        """Redraw the thing"""
        # Check if any of our viewer windows have closed
        for pid in self.pids[:]:
            (p, s) = os.waitpid(pid, os.WNOHANG)
            if p:
                self.pids.remove(pid)
            
        self.unTag()

        names = [x for x in self.a.submissions]                 
        names.sort()

        for name in names:
            sub = self.a.submissions[name]
            
            self.pathAdd([name])

            # If there are multiple usernames:
            if len(sub.getUsernames()) > 1:
                self.pathAdd([name, "Usernames"])
                for alt in sub.getUsernames():
                    self.pathAdd([name, "Usernames", alt])

            # Note the files
            self.pathAdd([name, "Files"])
            for f in sub.getFiles():
                if f is not sub.executable:
                    self.pathAdd([name, "Files", os.path.basename(f)])

            # Note the comments
            self.pathAdd([name, "Auto Comments"])
            self.pathAdd([name, "User Comments"])

            for c in sub.comments:
                if c.auto:
                    self.pathAdd([name, "Auto Comments",
                                  c.test.name + "(%s %f)" % (c.op, c.value)],
                                 data=c)
                else:
                    cName = None
                    for uc in self.a.userComments:
                        if self.a.userComments[uc] is c:
                            cName = uc
                            break

                    assert(cName is not None)
                    
                    self.pathAdd([name, "User Comments",
                                  cName], data=c)
        
        self.prune()
        MemTree.update(self)

    def onDClick(self, event):
        """Allow double-clicking, which depends on what was double
        clicked."""
        # Figure out what is selected.
        selId = self.GetSelection()
        parentId = self.GetItemParent(selId)
        
        # If it is a username, and we can do annotations, do that.
        if str(self.GetItemText(selId)) == self.getUsername():
            if self.parent.aFrame == None:
                self.parent.aFrame = AnnotateFrame(self.parent)
                self.parent.aFrame.Show()

        # If it is a filename, call the viewer for it.
        if parentId.IsOk() and str(self.GetItemText(parentId)) == "Files":
            self.onOpenFile()

        # If it is a comment, bring up the editor on it.
        if parentId.IsOk() and \
           (str(self.GetItemText(parentId)) == "User Comments" or \
            str(self.GetItemText(parentId)) == "Auto Comments"):
            self.onEditComment()

    def onKeyDown(self, event):
        # Trying to delete something, lets figure out what
        if event.GetKeyCode() == 127:
            # Figure out what is selected.
            selId = self.GetSelection()
            parentId = self.GetItemParent(selId)

            if not parentId.IsOk():
                return

            # If it is a submission, delete it
            if str(self.GetItemText(parentId)) == "Submissions":
                self.onDelSub()
                
            # If it is a filename, call the viewer for it.
            if str(self.GetItemText(parentId)) == "Files":
                self.onDelFile()

            # If it is a comment, bring up the editor on it.
            if (str(self.GetItemText(parentId)) == "User Comments" or \
                str(self.GetItemText(parentId)) == "Auto Comments"):
                self.onDelComment()
            

    def onRightUp(self, event):
        """Show the context menu"""
        id = self.HitTest(event.GetPosition())[0]
        if id.IsOk():
            self.SelectItem(id)

        menu = wx.Menu()

        menu.Append(self.contextNewSubID, "Add Submission")
        menu.Append(self.contextDelSubID, "Remove Submission")
        menu.AppendSeparator()

        selId = self.GetSelection()
        
        # For all submissions
        if selId != self.tree.id:            
            menu.Append(self.contextNewFileID, "Add File")
            menu.Append(self.contextDelFileID, "Remove File")
            menu.Append(self.contextNewUsernameID, "Add username")
            menu.Append(self.contextRunTestsID, "Run tests")

        # For files only
        if str(self.GetItemText(self.GetItemParent(selId))) == "Files":
            menu.AppendSeparator()
            menu.Append(self.contextOpenFileID, "Open File")

        # For comments only
        if str(self.GetItemText(self.GetItemParent(selId))) in \
           ["Auto Comments", "User Comments"]:
            menu.AppendSeparator()                        
            menu.Append(self.contextDelCommentID, "Unassign Comment")
            menu.Append(self.contextEditCommentID, "Edit Comment")

        # For usernames only
        if str(self.GetItemText(self.GetItemParent(selId))) == "Usernames":
            menu.AppendSeparator()
            menu.Append(self.contextDelUsernameID, "Remove Username")

        self.PopupMenu(menu, event.GetPosition())
        menu.Destroy()

    def onNewSub(self, event):
        # Check the base directory for directories
        dirNames = [os.path.basename(x) \
                    for x in os.listdir(self.a.submissions.basedir) \
                    if os.path.isdir(self.a.submissions.basedir + "/" + x)]

        # Find the ones not already listed as subs
        newNames = [x for x in dirNames if x not in self.a.submissions]

        # Allow the user to add any of those
        dlg = wx.MultiChoiceDialog(self,
                                   "Add which subdir(s) as new submissions?",
                                   "Add Submissions",
                                   choices=newNames)
        ret = dlg.ShowModal()
        indList = dlg.GetSelections()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        # These are the subs to add
        toAdd = [newNames[x] for x in indList]

        # Add 'em
        for newName in toAdd:
            try:
                self.a.submissions.add(newName)
            except:
                dlg = wx.MessageDialog(self,
                                       "Couldn't find files for " + newName,
                                       "Error",
                                       style=wx.OK|wx.CANCEL)
                ret = dlg.ShowModal()
                dlg.Destroy()

                if ret == wx.ID_CANCEL:
                    break                
                
        self.update()

    def onDelSub(self, event=None):
        """Delete a submission, after confirmation."""
        dlg = wx.MessageDialog(self,
                               "Are you sure you want to delete this " + \
                               "submission?",
                               "Confirm delete",
                               style=wx.YES_NO|wx.ICON_QUESTION)
        ret = dlg.ShowModal()
        dlg.Destroy()

        if ret != wx.ID_YES:
            return

        # Delete it
        name = self.getUsername()
        del self.a.submissions[name]

        self.update()

    def onNewFile(self, event):
        """Find the directories in the current directory that are not
        being used as submissions yet.  Pick from those which ones
        should be added.  Add the default files for those that are
        picked."""
        # Check the directory with the rest of this sub's files for
        # files that aren't in the list.
        
        sub = self.a.submissions[self.getUsername()]
        oldFiles = sub.getFiles()
        path = sub.getPath()

        choices = [os.path.basename(x) for x in os.listdir(path) \
                   if x not in oldFiles]

        dlg = wx.MultiChoiceDialog(self,
                                   "Add which file(s) to this submission?",
                                   "Add files",
                                   choices=choices)
        ret = dlg.ShowModal()
        indList = dlg.GetSelections()
        dlg.Destroy()

        fullPaths = [x for x in os.listdir(path) if x not in oldFiles]
        newFiles = [fullPaths[x] for x in indList]

        sub.files += newFiles

        self.update()

    def onDelFile(self, event=None):
        """Remove a file from a submission."""
        # Remove this file from the list
        name = str(self.GetItemText(self.GetSelection()))

        sub = self.a.submissions[self.getUsername()]

        if len(sub.files) <= 1:
            dlg = wx.MessageDialog(self,                  
                                   "Removing the last file in a submission "+\
                                   "removes the submission.  Continue?",
                                   "Remove sub?",
                                   style=wx.YES_NO|wx.ICON_QUESTION)
            ret = dlg.ShowModal()
            dlg.Destroy()
            if ret == wx.ID_YES:
                del self.a.submissions[self.getUsername()]
            return

        path = sub.getPath() + "/"

        sub.files.remove(path + name)                 

        self.update()

    def viewFile(self, cmd, filename):
        """Start up the viewer for a file."""
        pid = os.fork()
        if not pid:
            try:
                os.execvp(cmd, [cmd, filename])
            except:
                sys.exit("Error opening %s with %s" % (filename, cmd))
        else:
            self.pids.append(pid)

    def onOpenFile(self, event=None):
        """Start up the viewer for a file"""
        filename = str(self.GetItemText(self.GetSelection()))
        ext = filename.lower().rsplit(".", 1)[-1]

        if ext in self.viewers:
            cmd = self.viewers[ext][0]
        else:
            cmd = self.defaultViewer

        sub = self.a.submissions[self.getUsername()]
        for f in sub.files:
            if f.endswith("/" + filename):
                filename = f
                break
        
        self.viewFile(cmd, filename)

    def onRunTests(self, event):
        """Run tests on this submission."""
        self.progDlg = None
        try:
            self.a.runTestsOnSub(self.a.submissions[self.getUsername()],
                                 self.parent.runCallback)
        except KeyboardInterrupt:
            return
        except AssertionError:
            return
        except KeyError:
            return

        self.update()
        self.parent.resultTree.update()

    def onEditComment(self, event=None):
        """Edit a comment.  Supports both auto-comments and user comments."""
        c = self.getDataById(self.GetSelection())
        assert(c is not None)

        topNames = [x.name for x in self.a.rubric.getTopLevel()]
        commentNames = self.a.userComments.keys()
        selectedName = str(self.GetItemText(self.GetSelection()))

        if c.auto:
            dlg = CommentDialog(topNames, [], c, False,
                                "Auto Comment", self)
        else:
            dlg = CommentDialog(topNames, commentNames, c, False,
                                selectedName, self)
            
        ret = dlg.ShowModal()
        testName = dlg.rubricChoice.GetStringSelection()
        name = dlg.name
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        if not c.auto and name != selectedName:
            del self.a.userComments[selectedName]
            self.a.userComments[name] = c

        c.test = self.a.rubric[testName]

        self.update()        
        self.parent.commentTree.update()

    def onDelComment(self, event=None):
        """Remove a comment"""
        # Identify this comment and remove it
        c = self.getDataById(self.GetSelection())
        assert(c is not None)

        sub = self.a.submissions[self.getUsername()]

        sub.removeComment(c)

        self.update()

    def onNewUsername(self, event):
        """Add an extra username for this submission"""
        dlg = wx.TextEntryDialog(self, "New Username",
                                 style=wx.OK|wx.CANCEL)
        ret = dlg.ShowModal()
        val = dlg.GetValue()
        dlg.Destroy()

        if ret == wx.ID_CANCEL:
            return

        if not val:
            dlg = wx.MessageDialog(self, "Error",
                                   "You must enter a username",
                                   style=wx.OK|wx.CANCEL)
            dlg.ShowModal()
            dlg.Destroy()
            return

        # Get the sub
        sub = self.a.submissions[self.getUsername()]

        sub.addUsername(val)

        self.update()
        
    def onDelUsername(self, event):
        """Remove a username for this submission."""
        name = str(self.GetItemText(self.GetSelection()))
        sub = self.a.submissions[self.getUsername()]

        if name == self.getUsername():
            dlg = wx.MessageDialog(self,
                                   "You cannot remove the primary username",
                                   "Error",
                                   style=wx.OK|wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy
            return

        sub.removeUsername(name)

        self.update()
    
